@@ -8,12 +8,12 @@
  * vacuumlazy.c and analyze.c for the rest of the code for the latter two.
  *
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/commands/vacuum.c,v 1.244 2002/10/31 19:25:29 tgl Exp $
+ *	  $Header: /cvsroot/pgsql/src/backend/commands/vacuum.c,v 1.263.2.1 2004/12/23 22:42:32 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -166,7 +166,7 @@ vacuum(VacuumStmt *vacstmt)
 	if (vacstmt->verbose)
 		elevel = INFO;
 	else
-		elevel = DEBUG1;
+		elevel = DEBUG2;
 
 	/*
 	 * We cannot run VACUUM inside a user transaction block; if we were
@@ -189,11 +189,11 @@ vacuum(VacuumStmt *vacstmt)
 	/*
 	 * Create special memory context for cross-transaction storage.
 	 *
-	 * Since it is a child of QueryContext, it will go away eventually even
+	 * Since it is a child of PortalContext, it will go away eventually even
 	 * if we suffer an error; there's no need for special abort cleanup
 	 * logic.
 	 */
-	vac_context = AllocSetContextCreate(QueryContext,
+	vac_context = AllocSetContextCreate(PortalContext,
 										"Vacuum",
 										ALLOCSET_DEFAULT_MINSIZE,
 										ALLOCSET_DEFAULT_INITSIZE,
@@ -205,7 +205,7 @@ vacuum(VacuumStmt *vacstmt)
 	 * lifetime.
 	 */
 	if (vacstmt->analyze && !vacstmt->vacuum)
-		anl_context = AllocSetContextCreate(QueryContext,
+		anl_context = AllocSetContextCreate(PortalContext,
 											"Analyze",
 											ALLOCSET_DEFAULT_MINSIZE,
 											ALLOCSET_DEFAULT_INITSIZE,
@@ -275,7 +275,7 @@ vacuum(VacuumStmt *vacstmt)
 		}
 
 		/* matches the StartTransaction in PostgresMain() */
-		CommitTransactionCommand(true);
+		CommitTransactionCommand();
 	}
 
 	/*
@@ -283,28 +283,29 @@ vacuum(VacuumStmt *vacstmt)
 	 */
 	foreach(cur, vrl)
 	{
-		Oid			relid = (Oid) lfirsti(cur);
+		Oid			relid = lfirsto(cur);
 
 		if (vacstmt->vacuum)
 		{
-			if (! vacuum_rel(relid, vacstmt, RELKIND_RELATION))
-				all_rels = false; /* forget about updating dbstats */
+			if (!vacuum_rel(relid, vacstmt, RELKIND_RELATION))
+				all_rels = false;		/* forget about updating dbstats */
 		}
 		if (vacstmt->analyze)
 		{
 			MemoryContext old_context = NULL;
 
 			/*
-			 * If we vacuumed, use new transaction for analyze.
-			 * Otherwise, we can use the outer transaction, but we still
-			 * need to call analyze_rel in a memory context that will be
-			 * cleaned up on return (else we leak memory while processing
-			 * multiple tables).
+			 * If we vacuumed, use new transaction for analyze. Otherwise,
+			 * we can use the outer transaction, but we still need to call
+			 * analyze_rel in a memory context that will be cleaned up on
+			 * return (else we leak memory while processing multiple
+			 * tables).
 			 */
 			if (vacstmt->vacuum)
 			{
-				StartTransactionCommand(true);
-				SetQuerySnapshot();	/* might be needed for functional index */
+				StartTransactionCommand();
+				SetQuerySnapshot();		/* might be needed for functions
+										 * in indexes */
 			}
 			else
 				old_context = MemoryContextSwitchTo(anl_context);
@@ -312,7 +313,7 @@ vacuum(VacuumStmt *vacstmt)
 			analyze_rel(relid, vacstmt);
 
 			if (vacstmt->vacuum)
-				CommitTransactionCommand(true);
+				CommitTransactionCommand();
 			else
 			{
 				MemoryContextSwitchTo(old_context);
@@ -330,11 +331,16 @@ vacuum(VacuumStmt *vacstmt)
 
 		/*
 		 * This matches the CommitTransaction waiting for us in
-		 * PostgresMain(). We tell xact.c not to chain the upcoming
-		 * commit, so that a VACUUM doesn't start a transaction block,
-		 * even when autocommit is off.
+		 * PostgresMain().
 		 */
-		StartTransactionCommand(true);
+		StartTransactionCommand();
+
+		/*
+		 * If it was a database-wide VACUUM, print FSM usage statistics
+		 * (we don't make you be superuser to see these).
+		 */
+		if (vacstmt->relation == NULL)
+			PrintFreeSpaceMapStatistics(elevel);
 
 		/*
 		 * If we completed a database-wide VACUUM without skipping any
@@ -382,7 +388,7 @@ getrels(const RangeVar *vacrel, const char *stmttype)
 
 		/* Make a relation list entry for this guy */
 		oldcontext = MemoryContextSwitchTo(vac_context);
-		vrl = lappendi(vrl, relid);
+		vrl = lappendo(vrl, relid);
 		MemoryContextSwitchTo(oldcontext);
 	}
 	else
@@ -406,7 +412,7 @@ getrels(const RangeVar *vacrel, const char *stmttype)
 		{
 			/* Make a relation list entry for this guy */
 			oldcontext = MemoryContextSwitchTo(vac_context);
-			vrl = lappendi(vrl, HeapTupleGetOid(tuple));
+			vrl = lappendo(vrl, HeapTupleGetOid(tuple));
 			MemoryContextSwitchTo(oldcontext);
 		}
 
@@ -456,7 +462,9 @@ vacuum_set_xid_limits(VacuumStmt *vacstmt, bool sharedRel,
 	 */
 	if (TransactionIdFollows(limit, *oldestXmin))
 	{
-		elog(WARNING, "oldest Xmin is far in the past --- close open transactions soon to avoid wraparound problems");
+		ereport(WARNING,
+				(errmsg("oldest xmin is far in the past"),
+				 errhint("Close open transactions soon to avoid wraparound problems.")));
 		limit = *oldestXmin;
 	}
 
@@ -512,6 +520,9 @@ vac_update_relstats(Oid relid, BlockNumber num_pages, double num_tuples,
 		elog(ERROR, "pg_class entry for relid %u vanished during vacuuming",
 			 relid);
 
+	/* ensure no one else does this at the same time */
+	LockBuffer(buffer, BUFFER_LOCK_EXCLUSIVE);
+
 	/* overwrite the existing statistics in the tuple */
 	pgcform = (Form_pg_class) GETSTRUCT(&rtup);
 	pgcform->relpages = (int32) num_pages;
@@ -525,6 +536,8 @@ vac_update_relstats(Oid relid, BlockNumber num_pages, double num_tuples,
 	if (!hasindex)
 		pgcform->relhaspkey = false;
 
+	LockBuffer(buffer, BUFFER_LOCK_UNLOCK);
+
 	/*
 	 * Invalidate the tuple in the catcaches; this also arranges to flush
 	 * the relation's relcache entry.  (If we fail to commit for some
@@ -578,7 +591,10 @@ vac_update_dbstats(Oid dbid,
 	tuple = heap_getnext(scan, ForwardScanDirection);
 
 	if (!HeapTupleIsValid(tuple))
-		elog(ERROR, "database %u does not exist", dbid);
+		elog(ERROR, "could not find tuple for database %u", dbid);
+
+	/* ensure no one else does this at the same time */
+	LockBuffer(scan->rs_cbuf, BUFFER_LOCK_EXCLUSIVE);
 
 	dbform = (Form_pg_database) GETSTRUCT(tuple);
 
@@ -586,6 +602,8 @@ vac_update_dbstats(Oid dbid,
 	dbform->datvacuumxid = vacuumXID;
 	dbform->datfrozenxid = frozenXID;
 
+	LockBuffer(scan->rs_cbuf, BUFFER_LOCK_UNLOCK);
+
 	/* invalidate the tuple in the cache and write the buffer */
 	CacheInvalidateHeapTuple(relation, tuple);
 	WriteNoReleaseBuffer(scan->rs_cbuf);
@@ -662,8 +680,9 @@ vac_truncate_clog(TransactionId vacuumXID, TransactionId frozenXID)
 	 */
 	if (vacuumAlreadyWrapped)
 	{
-		elog(WARNING, "Some databases have not been vacuumed in over 2 billion transactions."
-			 "\n\tYou may have already suffered transaction-wraparound data loss.");
+		ereport(WARNING,
+				(errmsg("some databases have not been vacuumed in over 2 billion transactions"),
+				 errdetail("You may have already suffered transaction-wraparound data loss.")));
 		return;
 	}
 
@@ -673,17 +692,20 @@ vac_truncate_clog(TransactionId vacuumXID, TransactionId frozenXID)
 	/* Give warning about impending wraparound problems */
 	if (frozenAlreadyWrapped)
 	{
-		elog(WARNING, "Some databases have not been vacuumed in over 1 billion transactions."
-			 "\n\tBetter vacuum them soon, or you may have a wraparound failure.");
+		ereport(WARNING,
+				(errmsg("some databases have not been vacuumed in over 1 billion transactions"),
+				 errhint("Better vacuum them soon, or you may have a wraparound failure.")));
 	}
 	else
 	{
 		age = (int32) (myXID - frozenXID);
 		if (age > (int32) ((MaxTransactionId >> 3) * 3))
-			elog(WARNING, "Some databases have not been vacuumed in %d transactions."
-				 "\n\tBetter vacuum them within %d transactions,"
-				 "\n\tor you may have a wraparound failure.",
-				 age, (int32) (MaxTransactionId >> 1) - age);
+			ereport(WARNING,
+					(errmsg("some databases have not been vacuumed in %d transactions",
+							age),
+					 errhint("Better vacuum them within %d transactions, "
+							 "or you may have a wraparound failure.",
+							 (int32) (MaxTransactionId >> 1) - age)));
 	}
 }
 
@@ -722,8 +744,9 @@ vacuum_rel(Oid relid, VacuumStmt *vacstmt, char expected_relkind)
 	bool		result;
 
 	/* Begin a transaction for vacuuming this relation */
-	StartTransactionCommand(true);
-	SetQuerySnapshot();			/* might be needed for functional index */
+	StartTransactionCommand();
+	SetQuerySnapshot();			/* might be needed for functions in
+								 * indexes */
 
 	/*
 	 * Check for user-requested abort.	Note we want this to be inside a
@@ -739,7 +762,7 @@ vacuum_rel(Oid relid, VacuumStmt *vacstmt, char expected_relkind)
 							  ObjectIdGetDatum(relid),
 							  0, 0, 0))
 	{
-		CommitTransactionCommand(true);
+		CommitTransactionCommand();
 		return true;			/* okay 'cause no data there */
 	}
 
@@ -766,12 +789,13 @@ vacuum_rel(Oid relid, VacuumStmt *vacstmt, char expected_relkind)
 	onerel = relation_open(relid, lmode);
 
 	if (!(pg_class_ownercheck(RelationGetRelid(onerel), GetUserId()) ||
-		  (is_dbadmin(MyDatabaseId) && !onerel->rd_rel->relisshared)))
+		  (pg_database_ownercheck(MyDatabaseId, GetUserId()) && !onerel->rd_rel->relisshared)))
 	{
-		elog(WARNING, "Skipping \"%s\" --- only table or database owner can VACUUM it",
-			 RelationGetRelationName(onerel));
+		ereport(WARNING,
+				(errmsg("skipping \"%s\" --- only table or database owner can vacuum it",
+						RelationGetRelationName(onerel))));
 		relation_close(onerel, lmode);
-		CommitTransactionCommand(true);
+		CommitTransactionCommand();
 		return false;
 	}
 
@@ -781,10 +805,11 @@ vacuum_rel(Oid relid, VacuumStmt *vacstmt, char expected_relkind)
 	 */
 	if (onerel->rd_rel->relkind != expected_relkind)
 	{
-		elog(WARNING, "Skipping \"%s\" --- can not process indexes, views or special system tables",
-			 RelationGetRelationName(onerel));
+		ereport(WARNING,
+				(errmsg("skipping \"%s\" --- cannot vacuum indexes, views, or special system tables",
+						RelationGetRelationName(onerel))));
 		relation_close(onerel, lmode);
-		CommitTransactionCommand(true);
+		CommitTransactionCommand();
 		return false;
 	}
 
@@ -798,8 +823,9 @@ vacuum_rel(Oid relid, VacuumStmt *vacstmt, char expected_relkind)
 	if (isOtherTempNamespace(RelationGetNamespace(onerel)))
 	{
 		relation_close(onerel, lmode);
-		CommitTransactionCommand(true);
-		return true;			/* assume no long-lived data in temp tables */
+		CommitTransactionCommand();
+		return true;			/* assume no long-lived data in temp
+								 * tables */
 	}
 
 	/*
@@ -836,7 +862,7 @@ vacuum_rel(Oid relid, VacuumStmt *vacstmt, char expected_relkind)
 	/*
 	 * Complete the transaction and free all temporary memory used.
 	 */
-	CommitTransactionCommand(true);
+	CommitTransactionCommand();
 
 	/*
 	 * If the relation has a secondary toast rel, vacuum that too while we
@@ -847,7 +873,7 @@ vacuum_rel(Oid relid, VacuumStmt *vacstmt, char expected_relkind)
 	 */
 	if (toast_relid != InvalidOid)
 	{
-		if (! vacuum_rel(toast_relid, vacstmt, RELKIND_TOASTVALUE))
+		if (!vacuum_rel(toast_relid, vacstmt, RELKIND_TOASTVALUE))
 			result = false;		/* failed to vacuum the TOAST table? */
 	}
 
@@ -888,11 +914,6 @@ full_vacuum_rel(Relation onerel, VacuumStmt *vacstmt)
 	int			nindexes,
 				i;
 	VRelStats  *vacrelstats;
-	bool		reindex = false;
-
-	if (IsIgnoringSystemIndexes() &&
-		IsSystemRelation(onerel))
-		reindex = true;
 
 	vacuum_set_xid_limits(vacstmt, onerel->rd_rel->relisshared,
 						  &OldestXmin, &FreezeLimit);
@@ -911,27 +932,9 @@ full_vacuum_rel(Relation onerel, VacuumStmt *vacstmt)
 
 	/* Now open all indexes of the relation */
 	vac_open_indexes(onerel, &nindexes, &Irel);
-	if (!Irel)
-		reindex = false;
-	else if (!RelationGetForm(onerel)->relhasindex)
-		reindex = true;
 	if (nindexes > 0)
 		vacrelstats->hasindex = true;
 
-#ifdef NOT_USED
-
-	/*
-	 * reindex in VACUUM is dangerous under WAL. ifdef out until it
-	 * becomes safe.
-	 */
-	if (reindex)
-	{
-		vac_close_indexes(nindexes, Irel);
-		Irel = (Relation *) NULL;
-		activate_indexes_of_a_table(onerel, false);
-	}
-#endif   /* NOT_USED */
-
 	/* Clean/scan index relation(s) */
 	if (Irel != (Relation *) NULL)
 	{
@@ -974,16 +977,10 @@ full_vacuum_rel(Relation onerel, VacuumStmt *vacstmt)
 			 */
 			i = FlushRelationBuffers(onerel, vacrelstats->rel_pages);
 			if (i < 0)
-				elog(ERROR, "VACUUM (full_vacuum_rel): FlushRelationBuffers returned %d",
-					 i);
+				elog(ERROR, "FlushRelationBuffers returned %d", i);
 		}
 	}
 
-#ifdef NOT_USED
-	if (reindex)
-		activate_indexes_of_a_table(onerel, true);
-#endif   /* NOT_USED */
-
 	/* update shared free space map with final free space info */
 	vac_update_fsm(onerel, &fraged_pages, vacrelstats->rel_pages);
 
@@ -1020,15 +1017,13 @@ scan_heap(VRelStats *vacrelstats, Relation onerel,
 	VacPage		vacpage,
 				vacpagecopy;
 	BlockNumber empty_pages,
-				new_pages,
-				changed_pages,
 				empty_end_pages;
 	double		num_tuples,
 				tups_vacuumed,
 				nkeep,
 				nunused;
-	double		free_size,
-				usable_free_size;
+	double		free_space,
+				usable_free_space;
 	Size		min_tlen = MaxTupleSize;
 	Size		max_tlen = 0;
 	int			i;
@@ -1041,13 +1036,14 @@ scan_heap(VRelStats *vacrelstats, Relation onerel,
 	vac_init_rusage(&ru0);
 
 	relname = RelationGetRelationName(onerel);
-	elog(elevel, "--Relation %s.%s--",
-		 get_namespace_name(RelationGetNamespace(onerel)),
-		 relname);
+	ereport(elevel,
+			(errmsg("vacuuming \"%s.%s\"",
+					get_namespace_name(RelationGetNamespace(onerel)),
+					relname)));
 
-	empty_pages = new_pages = changed_pages = empty_end_pages = 0;
+	empty_pages = empty_end_pages = 0;
 	num_tuples = tups_vacuumed = nkeep = nunused = 0;
-	free_size = 0;
+	free_space = 0;
 
 	nblocks = RelationGetNumberOfBlocks(onerel);
 
@@ -1075,12 +1071,13 @@ scan_heap(VRelStats *vacrelstats, Relation onerel,
 
 		if (PageIsNew(page))
 		{
-			elog(WARNING, "Rel %s: Uninitialized page %u - fixing",
-				 relname, blkno);
+			ereport(WARNING,
+			(errmsg("relation \"%s\" page %u is uninitialized --- fixing",
+					relname, blkno)));
 			PageInit(page, BufferGetPageSize(buf), 0);
 			vacpage->free = ((PageHeader) page)->pd_upper - ((PageHeader) page)->pd_lower;
-			free_size += vacpage->free;
-			new_pages++;
+			free_space += vacpage->free;
+			empty_pages++;
 			empty_end_pages++;
 			vacpagecopy = copy_vac_page(vacpage);
 			vpage_insert(vacuum_pages, vacpagecopy);
@@ -1092,7 +1089,7 @@ scan_heap(VRelStats *vacrelstats, Relation onerel,
 		if (PageIsEmpty(page))
 		{
 			vacpage->free = ((PageHeader) page)->pd_upper - ((PageHeader) page)->pd_lower;
-			free_size += vacpage->free;
+			free_space += vacpage->free;
 			empty_pages++;
 			empty_end_pages++;
 			vacpagecopy = copy_vac_page(vacpage);
@@ -1188,9 +1185,12 @@ scan_heap(VRelStats *vacrelstats, Relation onerel,
 					/*
 					 * This should not happen, since we hold exclusive
 					 * lock on the relation; shouldn't we raise an error?
+					 * (Actually, it can happen in system catalogs, since
+					 * we tend to release write lock before commit there.)
 					 */
-					elog(WARNING, "Rel %s: TID %u/%u: InsertTransactionInProgress %u - can't shrink relation",
-						 relname, blkno, offnum, HeapTupleHeaderGetXmin(tuple.t_data));
+					ereport(NOTICE,
+							(errmsg("relation \"%s\" TID %u/%u: InsertTransactionInProgress %u --- can't shrink relation",
+									relname, blkno, offnum, HeapTupleHeaderGetXmin(tuple.t_data))));
 					do_shrinking = false;
 					break;
 				case HEAPTUPLE_DELETE_IN_PROGRESS:
@@ -1198,13 +1198,16 @@ scan_heap(VRelStats *vacrelstats, Relation onerel,
 					/*
 					 * This should not happen, since we hold exclusive
 					 * lock on the relation; shouldn't we raise an error?
+					 * (Actually, it can happen in system catalogs, since
+					 * we tend to release write lock before commit there.)
 					 */
-					elog(WARNING, "Rel %s: TID %u/%u: DeleteTransactionInProgress %u - can't shrink relation",
-						 relname, blkno, offnum, HeapTupleHeaderGetXmax(tuple.t_data));
+					ereport(NOTICE,
+							(errmsg("relation \"%s\" TID %u/%u: DeleteTransactionInProgress %u --- can't shrink relation",
+									relname, blkno, offnum, HeapTupleHeaderGetXmax(tuple.t_data))));
 					do_shrinking = false;
 					break;
 				default:
-					elog(ERROR, "Unexpected HeapTupleSatisfiesVacuum result");
+					elog(ERROR, "unexpected HeapTupleSatisfiesVacuum result");
 					break;
 			}
 
@@ -1217,8 +1220,8 @@ scan_heap(VRelStats *vacrelstats, Relation onerel,
 			 */
 			if (onerel->rd_rel->relhasoids &&
 				!OidIsValid(HeapTupleGetOid(&tuple)))
-				elog(WARNING, "Rel %s: TID %u/%u: OID IS INVALID. TUPGONE %d.",
-					 relname, blkno, offnum, (int) tupgone);
+				elog(WARNING, "relation \"%s\" TID %u/%u: OID is invalid",
+					 relname, blkno, offnum);
 
 			if (tupgone)
 			{
@@ -1275,7 +1278,7 @@ scan_heap(VRelStats *vacrelstats, Relation onerel,
 			do_reap = (vacpage->offsets_free > 0);
 		}
 
-		free_size += vacpage->free;
+		free_space += vacpage->free;
 
 		/*
 		 * Add the page to fraged_pages if it has a useful amount of free
@@ -1294,16 +1297,21 @@ scan_heap(VRelStats *vacrelstats, Relation onerel,
 				vpage_insert(fraged_pages, vacpagecopy);
 		}
 
+		/*
+		 * Include the page in empty_end_pages if it will be empty after
+		 * vacuuming; this is to keep us from using it as a move
+		 * destination.
+		 */
 		if (notup)
+		{
+			empty_pages++;
 			empty_end_pages++;
+		}
 		else
 			empty_end_pages = 0;
 
 		if (pgchanged)
-		{
 			WriteBuffer(buf);
-			changed_pages++;
-		}
 		else
 			ReleaseBuffer(buf);
 	}
@@ -1330,14 +1338,14 @@ scan_heap(VRelStats *vacrelstats, Relation onerel,
 	{
 		Assert((BlockNumber) fraged_pages->num_pages >= empty_end_pages);
 		fraged_pages->num_pages -= empty_end_pages;
-		usable_free_size = 0;
+		usable_free_space = 0;
 		for (i = 0; i < fraged_pages->num_pages; i++)
-			usable_free_size += fraged_pages->pagedesc[i]->free;
+			usable_free_space += fraged_pages->pagedesc[i]->free;
 	}
 	else
 	{
 		fraged_pages->num_pages = 0;
-		usable_free_size = 0;
+		usable_free_space = 0;
 	}
 
 	/* don't bother to save vtlinks if we will not call repair_frag */
@@ -1355,17 +1363,24 @@ scan_heap(VRelStats *vacrelstats, Relation onerel,
 		pfree(vtlinks);
 	}
 
-	elog(elevel, "Pages %u: Changed %u, reaped %u, Empty %u, New %u; "
-		 "Tup %.0f: Vac %.0f, Keep/VTL %.0f/%u, UnUsed %.0f, MinLen %lu, "
-		 "MaxLen %lu; Re-using: Free/Avail. Space %.0f/%.0f; "
-		 "EndEmpty/Avail. Pages %u/%u.\n\t%s",
-		 nblocks, changed_pages, vacuum_pages->num_pages, empty_pages,
-		 new_pages, num_tuples, tups_vacuumed,
-		 nkeep, vacrelstats->num_vtlinks,
-		 nunused, (unsigned long) min_tlen, (unsigned long) max_tlen,
-		 free_size, usable_free_size,
-		 empty_end_pages, fraged_pages->num_pages,
-		 vac_show_rusage(&ru0));
+	ereport(elevel,
+			(errmsg("\"%s\": found %.0f removable, %.0f nonremovable row versions in %u pages",
+					RelationGetRelationName(onerel),
+					tups_vacuumed, num_tuples, nblocks),
+			 errdetail("%.0f dead row versions cannot be removed yet.\n"
+				"Nonremovable row versions range from %lu to %lu bytes long.\n"
+					   "There were %.0f unused item pointers.\n"
+		 "Total free space (including removable row versions) is %.0f bytes.\n"
+					   "%u pages are or will become empty, including %u at the end of the table.\n"
+					   "%u pages containing %.0f free bytes are potential move destinations.\n"
+					   "%s",
+					   nkeep,
+					   (unsigned long) min_tlen, (unsigned long) max_tlen,
+					   nunused,
+					   free_space,
+					   empty_pages, empty_end_pages,
+					   fraged_pages->num_pages, usable_free_space,
+					   vac_show_rusage(&ru0))));
 }
 
 
@@ -1437,6 +1452,8 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 	 * We need a ResultRelInfo and an EState so we can use the regular
 	 * executor's index-entry-making machinery.
 	 */
+	estate = CreateExecutorState();
+
 	resultRelInfo = makeNode(ResultRelInfo);
 	resultRelInfo->ri_RangeTableIndex = 1;		/* dummy */
 	resultRelInfo->ri_RelationDesc = onerel;
@@ -1444,7 +1461,6 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 
 	ExecOpenIndices(resultRelInfo);
 
-	estate = CreateExecutorState();
 	estate->es_result_relations = resultRelInfo;
 	estate->es_num_result_relations = 1;
 	estate->es_result_relation_info = resultRelInfo;
@@ -1588,7 +1604,7 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 				if (tuple.t_data->t_infomask & HEAP_MOVED_OFF)
 				{
 					if (HeapTupleHeaderGetXvac(tuple.t_data) != myXID)
-						elog(ERROR, "Invalid XVAC in tuple header");
+						elog(ERROR, "invalid XVAC in tuple header");
 					if (keep_tuples == 0)
 						continue;
 					if (chain_tuple_moved)		/* some chains was moved
@@ -1667,7 +1683,7 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 				/* Quick exit if we have no vtlinks to search in */
 				if (vacrelstats->vtlinks == NULL)
 				{
-					elog(DEBUG1, "Parent item in update-chain not found - can't continue repair_frag");
+					elog(DEBUG2, "parent item in update-chain not found --- can't continue repair_frag");
 					break;		/* out of walk-along-page loop */
 				}
 
@@ -1704,7 +1720,7 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 						 * in scan_heap(), but it's not implemented at the
 						 * moment and so we just stop shrinking here.
 						 */
-						elog(DEBUG1, "Child itemid in update-chain marked as unused - can't continue repair_frag");
+						elog(DEBUG2, "child itemid in update-chain marked as unused --- can't continue repair_frag");
 						chain_move_failed = true;
 						break;	/* out of loop to move to chain end */
 					}
@@ -1789,7 +1805,7 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 					if (vtlp == NULL)
 					{
 						/* see discussion above */
-						elog(DEBUG1, "Parent item in update-chain not found - can't continue repair_frag");
+						elog(DEBUG2, "parent item in update-chain not found --- can't continue repair_frag");
 						chain_move_failed = true;
 						break;	/* out of check-all-items loop */
 					}
@@ -1801,7 +1817,7 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 							   ItemPointerGetOffsetNumber(&(tp.t_self)));
 					/* this can't happen since we saw tuple earlier: */
 					if (!ItemIdIsUsed(Pitemid))
-						elog(ERROR, "Parent itemid marked as unused");
+						elog(ERROR, "parent itemid marked as unused");
 					Ptp.t_datamcxt = NULL;
 					Ptp.t_data = (HeapTupleHeader) PageGetItem(Ppage, Pitemid);
 
@@ -1825,7 +1841,7 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 									 HeapTupleHeaderGetXmin(tp.t_data))))
 					{
 						ReleaseBuffer(Pbuf);
-						elog(DEBUG1, "Too old parent tuple found - can't continue repair_frag");
+						elog(DEBUG2, "too old parent tuple found --- can't continue repair_frag");
 						chain_move_failed = true;
 						break;	/* out of check-all-items loop */
 					}
@@ -1898,7 +1914,7 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 					 */
 					CacheInvalidateHeapTuple(onerel, &tuple);
 
-					/* NO ELOG(ERROR) TILL CHANGES ARE LOGGED */
+					/* NO EREPORT(ERROR) TILL CHANGES ARE LOGGED */
 					START_CRIT_SECTION();
 
 					tuple.t_data->t_infomask &= ~(HEAP_XMIN_COMMITTED |
@@ -1954,7 +1970,7 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 										 LP_USED);
 					if (newoff == InvalidOffsetNumber)
 					{
-						elog(PANIC, "moving chain: failed to add item with len = %lu to page %u",
+						elog(PANIC, "failed to add item with len = %lu to page %u while moving tuple chain",
 						  (unsigned long) tuple_len, destvacpage->blkno);
 					}
 					newitemid = PageGetItemId(ToPage, newoff);
@@ -2025,6 +2041,7 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 						ExecStoreTuple(&newtup, slot, InvalidBuffer, false);
 						ExecInsertIndexTuples(slot, &(newtup.t_self),
 											  estate, true);
+						ResetPerTupleExprContext(estate);
 					}
 
 					WriteBuffer(cur_buffer);
@@ -2081,7 +2098,7 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 			 */
 			CacheInvalidateHeapTuple(onerel, &tuple);
 
-			/* NO ELOG(ERROR) TILL CHANGES ARE LOGGED */
+			/* NO EREPORT(ERROR) TILL CHANGES ARE LOGGED */
 			START_CRIT_SECTION();
 
 			/*
@@ -2158,6 +2175,7 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 			{
 				ExecStoreTuple(&newtup, slot, InvalidBuffer, false);
 				ExecInsertIndexTuples(slot, &(newtup.t_self), estate, true);
+				ResetPerTupleExprContext(estate);
 			}
 		}						/* walk along page */
 
@@ -2187,11 +2205,11 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 				if (tuple.t_data->t_infomask & HEAP_XMIN_COMMITTED)
 					continue;
 				if (tuple.t_data->t_infomask & HEAP_MOVED_IN)
-					elog(ERROR, "HEAP_MOVED_IN was not expected (2)");
+					elog(ERROR, "HEAP_MOVED_IN was not expected");
 				if (tuple.t_data->t_infomask & HEAP_MOVED_OFF)
 				{
 					if (HeapTupleHeaderGetXvac(tuple.t_data) != myXID)
-						elog(ERROR, "Invalid XVAC in tuple header (4)");
+						elog(ERROR, "invalid XVAC in tuple header");
 					/* some chains was moved while */
 					if (chain_tuple_moved)
 					{			/* cleaning this page */
@@ -2216,7 +2234,7 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 					}
 				}
 				else
-					elog(ERROR, "HEAP_MOVED_OFF was expected (2)");
+					elog(ERROR, "HEAP_MOVED_OFF was expected");
 			}
 		}
 
@@ -2293,7 +2311,7 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 	 * status bits.  This is not really necessary, but will save time for
 	 * future transactions examining these tuples.
 	 *
-	 * XXX WARNING that this code fails to clear HEAP_MOVED_OFF tuples from
+	 * XXX NOTICE that this code fails to clear HEAP_MOVED_OFF tuples from
 	 * pages that were move source pages but not move dest pages.  One
 	 * also wonders whether it wouldn't be better to skip this step and
 	 * let the tuple status updates happen someplace that's not holding an
@@ -2329,7 +2347,7 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 				if (!(tuple.t_data->t_infomask & HEAP_MOVED))
 					elog(ERROR, "HEAP_MOVED_OFF/HEAP_MOVED_IN was expected");
 				if (HeapTupleHeaderGetXvac(tuple.t_data) != myXID)
-					elog(ERROR, "Invalid XVAC in tuple header (2)");
+					elog(ERROR, "invalid XVAC in tuple header");
 				if (tuple.t_data->t_infomask & HEAP_MOVED_IN)
 				{
 					tuple.t_data->t_infomask |= HEAP_XMIN_COMMITTED;
@@ -2347,10 +2365,18 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 	}
 	Assert(num_moved == checked_moved);
 
-	elog(elevel, "Rel %s: Pages: %u --> %u; Tuple(s) moved: %u.\n\t%s",
-		 RelationGetRelationName(onerel),
-		 nblocks, blkno, num_moved,
-		 vac_show_rusage(&ru0));
+	/*
+	 * It'd be cleaner to make this report at the bottom of this routine,
+	 * but then the rusage would double-count the second pass of index
+	 * vacuuming.  So do it here and ignore the relatively small amount of
+	 * processing that occurs below.
+	 */
+	ereport(elevel,
+			(errmsg("\"%s\": moved %u row versions, truncated %u to %u pages",
+					RelationGetRelationName(onerel),
+					num_moved, nblocks, blkno),
+			 errdetail("%s",
+					   vac_show_rusage(&ru0))));
 
 	/*
 	 * Reflect the motion of system tuples to catalog cache here.
@@ -2385,8 +2411,7 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 		if (vacpage->blkno == (blkno - 1) &&
 			vacpage->offsets_free > 0)
 		{
-			OffsetNumber unbuf[BLCKSZ / sizeof(OffsetNumber)];
-			OffsetNumber *unused = unbuf;
+			OffsetNumber unused[BLCKSZ / sizeof(OffsetNumber)];
 			int			uncnt;
 
 			buf = ReadBuffer(onerel, vacpage->blkno);
@@ -2409,12 +2434,12 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 					if (tuple.t_data->t_infomask & HEAP_MOVED_OFF)
 					{
 						if (HeapTupleHeaderGetXvac(tuple.t_data) != myXID)
-							elog(ERROR, "Invalid XVAC in tuple header (3)");
+							elog(ERROR, "invalid XVAC in tuple header");
 						itemid->lp_flags &= ~LP_USED;
 						num_tuples++;
 					}
 					else
-						elog(ERROR, "HEAP_MOVED_OFF was expected (3)");
+						elog(ERROR, "HEAP_MOVED_OFF was expected");
 				}
 
 			}
@@ -2429,8 +2454,7 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 			{
 				XLogRecPtr	recptr;
 
-				recptr = log_heap_clean(onerel, buf, (char *) unused,
-						  (char *) (&(unused[uncnt])) - (char *) unused);
+				recptr = log_heap_clean(onerel, buf, unused, uncnt);
 				PageSetLSN(page, recptr);
 				PageSetSUI(page, ThisStartUpID);
 			}
@@ -2464,8 +2488,7 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 	 */
 	i = FlushRelationBuffers(onerel, blkno);
 	if (i < 0)
-		elog(ERROR, "VACUUM (repair_frag): FlushRelationBuffers returned %d",
-			 i);
+		elog(ERROR, "FlushRelationBuffers returned %d", i);
 
 	/* truncate relation, if needed */
 	if (blkno < nblocks)
@@ -2484,6 +2507,8 @@ repair_frag(VRelStats *vacrelstats, Relation onerel,
 	ExecDropTupleTable(tupleTable, true);
 
 	ExecCloseIndices(resultRelInfo);
+
+	FreeExecutorState(estate);
 }
 
 /*
@@ -2528,15 +2553,15 @@ vacuum_heap(VRelStats *vacrelstats, Relation onerel, VacPageList vacuum_pages)
 
 	i = FlushRelationBuffers(onerel, relblocks);
 	if (i < 0)
-		elog(ERROR, "VACUUM (vacuum_heap): FlushRelationBuffers returned %d",
-			 i);
+		elog(ERROR, "FlushRelationBuffers returned %d", i);
 
 	/* truncate relation if there are some empty end-pages */
 	if (vacuum_pages->empty_end_pages > 0)
 	{
-		elog(elevel, "Rel %s: Pages: %u --> %u.",
-			 RelationGetRelationName(onerel),
-			 vacrelstats->rel_pages, relblocks);
+		ereport(elevel,
+				(errmsg("\"%s\": truncated %u to %u pages",
+						RelationGetRelationName(onerel),
+						vacrelstats->rel_pages, relblocks)));
 		relblocks = smgrtruncate(DEFAULT_SMGR, onerel, relblocks);
 		onerel->rd_nblocks = relblocks; /* update relcache immediately */
 		onerel->rd_targblock = InvalidBlockNumber;
@@ -2552,8 +2577,7 @@ vacuum_heap(VRelStats *vacrelstats, Relation onerel, VacPageList vacuum_pages)
 static void
 vacuum_page(Relation onerel, Buffer buffer, VacPage vacpage)
 {
-	OffsetNumber unbuf[BLCKSZ / sizeof(OffsetNumber)];
-	OffsetNumber *unused = unbuf;
+	OffsetNumber unused[BLCKSZ / sizeof(OffsetNumber)];
 	int			uncnt;
 	Page		page = BufferGetPage(buffer);
 	ItemId		itemid;
@@ -2577,8 +2601,7 @@ vacuum_page(Relation onerel, Buffer buffer, VacPage vacpage)
 	{
 		XLogRecPtr	recptr;
 
-		recptr = log_heap_clean(onerel, buffer, (char *) unused,
-						  (char *) (&(unused[uncnt])) - (char *) unused);
+		recptr = log_heap_clean(onerel, buffer, unused, uncnt);
 		PageSetLSN(page, recptr);
 		PageSetSUI(page, ThisStartUpID);
 	}
@@ -2600,17 +2623,25 @@ static void
 scan_index(Relation indrel, double num_tuples)
 {
 	IndexBulkDeleteResult *stats;
+	IndexVacuumCleanupInfo vcinfo;
 	VacRUsage	ru0;
 
 	vac_init_rusage(&ru0);
 
 	/*
-	 * Even though we're not planning to delete anything, use the
-	 * ambulkdelete call, so that the scan happens within the index AM for
-	 * more speed.
+	 * Even though we're not planning to delete anything, we use the
+	 * ambulkdelete call, because (a) the scan happens within the index AM
+	 * for more speed, and (b) it may want to pass private statistics to
+	 * the amvacuumcleanup call.
 	 */
 	stats = index_bulk_delete(indrel, dummy_tid_reaped, NULL);
 
+	/* Do post-VACUUM cleanup, even though we deleted nothing */
+	vcinfo.vacuum_full = true;
+	vcinfo.message_level = elevel;
+
+	stats = index_vacuum_cleanup(indrel, &vcinfo, stats);
+
 	if (!stats)
 		return;
 
@@ -2619,10 +2650,15 @@ scan_index(Relation indrel, double num_tuples)
 						stats->num_pages, stats->num_index_tuples,
 						false);
 
-	elog(elevel, "Index %s: Pages %u; Tuples %.0f.\n\t%s",
-		 RelationGetRelationName(indrel),
-		 stats->num_pages, stats->num_index_tuples,
-		 vac_show_rusage(&ru0));
+	ereport(elevel,
+			(errmsg("index \"%s\" now contains %.0f row versions in %u pages",
+					RelationGetRelationName(indrel),
+					stats->num_index_tuples,
+					stats->num_pages),
+			 errdetail("%u index pages have been deleted, %u are currently reusable.\n"
+					   "%s",
+					   stats->pages_deleted, stats->pages_free,
+					   vac_show_rusage(&ru0))));
 
 	/*
 	 * Check for tuple count mismatch.	If the index is partial, then it's
@@ -2632,10 +2668,11 @@ scan_index(Relation indrel, double num_tuples)
 	{
 		if (stats->num_index_tuples > num_tuples ||
 			!vac_is_partial_index(indrel))
-			elog(WARNING, "Index %s: NUMBER OF INDEX' TUPLES (%.0f) IS NOT THE SAME AS HEAP' (%.0f)."
-				 "\n\tRecreate the index.",
-				 RelationGetRelationName(indrel),
-				 stats->num_index_tuples, num_tuples);
+			ereport(WARNING,
+					(errmsg("index \"%s\" contains %.0f row versions, but table contains %.0f row versions",
+							RelationGetRelationName(indrel),
+							stats->num_index_tuples, num_tuples),
+					 errhint("Rebuild the index with REINDEX.")));
 	}
 
 	pfree(stats);
@@ -2658,6 +2695,7 @@ vacuum_index(VacPageList vacpagelist, Relation indrel,
 			 double num_tuples, int keep_tuples)
 {
 	IndexBulkDeleteResult *stats;
+	IndexVacuumCleanupInfo vcinfo;
 	VacRUsage	ru0;
 
 	vac_init_rusage(&ru0);
@@ -2665,6 +2703,12 @@ vacuum_index(VacPageList vacpagelist, Relation indrel,
 	/* Do bulk deletion */
 	stats = index_bulk_delete(indrel, tid_reaped, (void *) vacpagelist);
 
+	/* Do post-VACUUM cleanup */
+	vcinfo.vacuum_full = true;
+	vcinfo.message_level = elevel;
+
+	stats = index_vacuum_cleanup(indrel, &vcinfo, stats);
+
 	if (!stats)
 		return;
 
@@ -2673,10 +2717,17 @@ vacuum_index(VacPageList vacpagelist, Relation indrel,
 						stats->num_pages, stats->num_index_tuples,
 						false);
 
-	elog(elevel, "Index %s: Pages %u; Tuples %.0f: Deleted %.0f.\n\t%s",
-		 RelationGetRelationName(indrel), stats->num_pages,
-		 stats->num_index_tuples - keep_tuples, stats->tuples_removed,
-		 vac_show_rusage(&ru0));
+	ereport(elevel,
+			(errmsg("index \"%s\" now contains %.0f row versions in %u pages",
+					RelationGetRelationName(indrel),
+					stats->num_index_tuples,
+					stats->num_pages),
+			 errdetail("%.0f index row versions were removed.\n"
+		 "%u index pages have been deleted, %u are currently reusable.\n"
+					   "%s",
+					   stats->tuples_removed,
+					   stats->pages_deleted, stats->pages_free,
+					   vac_show_rusage(&ru0))));
 
 	/*
 	 * Check for tuple count mismatch.	If the index is partial, then it's
@@ -2686,10 +2737,11 @@ vacuum_index(VacPageList vacpagelist, Relation indrel,
 	{
 		if (stats->num_index_tuples > num_tuples + keep_tuples ||
 			!vac_is_partial_index(indrel))
-			elog(WARNING, "Index %s: NUMBER OF INDEX' TUPLES (%.0f) IS NOT THE SAME AS HEAP' (%.0f)."
-				 "\n\tRecreate the index.",
-				 RelationGetRelationName(indrel),
-				 stats->num_index_tuples, num_tuples);
+			ereport(WARNING,
+					(errmsg("index \"%s\" contains %.0f row versions, but table contains %.0f row versions",
+							RelationGetRelationName(indrel),
+					  stats->num_index_tuples, num_tuples + keep_tuples),
+					 errhint("Rebuild the index with REINDEX.")));
 	}
 
 	pfree(stats);
@@ -2765,31 +2817,49 @@ vac_update_fsm(Relation onerel, VacPageList fraged_pages,
 			   BlockNumber rel_pages)
 {
 	int			nPages = fraged_pages->num_pages;
-	int			i;
+	VacPage    *pagedesc = fraged_pages->pagedesc;
+	Size		threshold;
 	PageFreeSpaceInfo *pageSpaces;
+	int			outPages;
+	int			i;
+
+	/*
+	 * We only report pages with free space at least equal to the average
+	 * request size --- this avoids cluttering FSM with uselessly-small
+	 * bits of space.  Although FSM would discard pages with little free
+	 * space anyway, it's important to do this prefiltering because (a) it
+	 * reduces the time spent holding the FSM lock in
+	 * RecordRelationFreeSpace, and (b) FSM uses the number of pages
+	 * reported as a statistic for guiding space management.  If we didn't
+	 * threshold our reports the same way vacuumlazy.c does, we'd be
+	 * skewing that statistic.
+	 */
+	threshold = GetAvgFSMRequestSize(&onerel->rd_node);
 
 	/* +1 to avoid palloc(0) */
 	pageSpaces = (PageFreeSpaceInfo *)
 		palloc((nPages + 1) * sizeof(PageFreeSpaceInfo));
+	outPages = 0;
 
 	for (i = 0; i < nPages; i++)
 	{
-		pageSpaces[i].blkno = fraged_pages->pagedesc[i]->blkno;
-		pageSpaces[i].avail = fraged_pages->pagedesc[i]->free;
-
 		/*
 		 * fraged_pages may contain entries for pages that we later
 		 * decided to truncate from the relation; don't enter them into
 		 * the free space map!
 		 */
-		if (pageSpaces[i].blkno >= rel_pages)
-		{
-			nPages = i;
+		if (pagedesc[i]->blkno >= rel_pages)
 			break;
+
+		if (pagedesc[i]->free >= threshold)
+		{
+			pageSpaces[outPages].blkno = pagedesc[i]->blkno;
+			pageSpaces[outPages].avail = pagedesc[i]->free;
+			outPages++;
 		}
 	}
 
-	MultiRecordFreeSpace(&onerel->rd_node, 0, nPages, pageSpaces);
+	RecordRelationFreeSpace(&onerel->rd_node, outPages, pageSpaces);
 
 	pfree(pageSpaces);
 }
@@ -2954,7 +3024,7 @@ vac_open_indexes(Relation relation, int *nindexes, Relation **Irel)
 	i = 0;
 	foreach(indexoidscan, indexoidlist)
 	{
-		Oid			indexoid = lfirsti(indexoidscan);
+		Oid			indexoid = lfirsto(indexoidscan);
 
 		(*Irel)[i] = index_open(indexoid);
 		i++;
@@ -2990,7 +3060,10 @@ vac_is_partial_index(Relation indrel)
 		return true;
 
 	/* Otherwise, look to see if there's a partial-index predicate */
-	return (VARSIZE(&indrel->rd_index->indpred) > VARHDRSZ);
+	if (!heap_attisnull(indrel->rd_indextuple, Anum_pg_index_indpred))
+		return true;
+
+	return false;
 }
 
 
